/*
 */

#include <cutil.h>
#include <cstdio>
//#include <cstdlib>
//#include <string.h>

// kernel routines
#include <nbody_kernel.cu> // included here! is not compiled separately!
  // this one does kick step and all pairs forces
#include <energy_kernel.cu> 
  // all pairs potential energy and kinetic energy computation
// #include <kep_kernel.cu> 
  // f+g function keplerian evolution on GPU for N particles
#include <drift_kernel.cu> 
  // drift step computation on GPU
  // also contains parallel reduction sum used in energysum
  // and m*v vector on GPU routine

extern "C"
{


void checkCUDA()
{   
    CUT_DEVICE_INIT();
}

// never used because last time I tried it
// the device only got 0
void setDeviceSoftening(float softening)
{
    float softeningSq = softening*softening;

    CUDA_SAFE_CALL(cudaMemcpyToSymbol("softeningSquared", &softeningSq, 
                                      sizeof(float), 0, 
                                      cudaMemcpyHostToDevice));
    printf("softeningSq=%f \n",softeningSq); // this works fine here
}

// memory for nbody system
void allocateNBodyArrays(float* vel[2], int numBodies)
{
    // 4 floats each for alignment reasons
    unsigned int memSize = sizeof( float) * 4 * numBodies;
    
    CUDA_SAFE_CALL(cudaMalloc((void**)&vel[0], memSize));
    CUDA_SAFE_CALL(cudaMalloc((void**)&vel[1], memSize));
}


void deleteNBodyArrays(float* vel[2])
{
    CUDA_SAFE_CALL(cudaFree((void**)vel[0]));
    CUDA_SAFE_CALL(cudaFree((void**)vel[1]));
}

// note times 4 in size
void copyArrayFromDevice(float* host, 
                         const float* device, 
                         int numBodies)
{   
    CUDA_SAFE_CALL(cudaMemcpy(host, device, numBodies*4*sizeof(float), 
                              cudaMemcpyDeviceToHost));
}

// note times 4 in size
void copyArrayToDevice(float* device, const float* host, int numBodies)
{
    CUDA_SAFE_CALL(cudaMemcpy(device, host, numBodies*4*sizeof(float), 
                              cudaMemcpyHostToDevice));
}


// note newPos and other arrays are recast here in this routine
// from float* to float4*, 
//   you cannot change to float4* in function  def
// the reason for this is that float4 is not defined in regular cpp or c
// but is defined for the .cu files

void threadSync() { cudaThreadSynchronize(); }

///////////////////////////////////////////////////////
// not used for Keplerian integration methods
void 
integrateNbodySystem(float* newPos, float* newVel, 
                     float* oldPos, float* oldVel, 
                     float deltaTime, float damping,  float soft2,
                     int numBodies, int p)
{
    static int firsttime = 0;
    int sharedMemSize = p *  sizeof(float4); // 4 floats for pos
    
    dim3 threads(p,1,1); // number of threads per block is p  blockdim.x=p
    dim3 grid(numBodies/p, 1, 1);
    // number of blocks is griddim.x=numBodies/p and gridDim.y=1
    if (firsttime ==0){
       printf("number blocks %d, number threads=%d\n",grid.x,threads.x);
       fflush(stdout);
       firsttime=1;
    }


    // execute the kernel: type conversion to float4 done here
    threadSync();
    integrateBodies<<< grid, threads, sharedMemSize >>>
            ((float4*)newPos, (float4*)newVel,
            (float4*)oldPos, (float4*)oldVel,
            deltaTime, damping, soft2, numBodies);
    

    // check if kernel invocation generated an error
    CUT_CHECK_ERROR("Nbody Kernel execution failed");
}

///////////////////////////////////////////////////////
// change the central mass in global device memory
// primarily used during energy computation
// other routines assume central mass set to zero previously
///////////////////////////////////////////////////////
void
callchangem0(float *aPos, float m0) 
{
    int sharedMemSize = 0;
    dim3 threads(1,1,1); 
    dim3 grid(1,1,1);
    changem0<<< grid, threads, sharedMemSize >>>((float4*)aPos, m0);
    CUT_CHECK_ERROR("change m0 Kernel execution failed");
}

// forward declare
void sumMomenta(float* Pos, float* Vel, int p, int numBodies, float *pxpypz);  

////////////////////////////////////////////////////////////
// compute the energyvector and then sum it
// if m0==0 then cartesian coordinates assumed
// if m0>0 then helio/bary coordinates assumed
// energyvector all pairs computed on GPU
// sums done with parallel reduction algorith
// note that energyvector done with m0!=0 if helio/bary
// V0 = Q0 = 0 should be set previously
////////////////////////////////////////////////////////////
float energysum(float* Pos,  float* Vel,
                     float soft2, int numBodies, int p, float m0)
{
    int sharedMemSize = p *  sizeof(float4); // 4 floats for pos
    dim3 threads(p,1,1); // number of threads per block is p  blockdim.x=p
    dim3 grid(numBodies/p, 1, 1);

    float *d_vec; // energy vector on device
    CUDA_SAFE_CALL( cudaMalloc( (void**) &d_vec, numBodies*sizeof(float)));

    if (m0 >0.0) callchangem0(Pos, m0);  
   // if helio/bary then reset central mass to non zero
   // this is so central mass interaction terms are added
   // as long as Q0 com term = 0 then the central mass interaction
   // terms should be correct
   // if we have set V0 to zero then it doesn't matter
   // that KE added in 0 term also

// computes a vector that has 0.5* Sum_i m_i *[ v_i^2 - Sum_{j!=i} m_j/r_ij ]
    BodyKEPE<<< grid, threads, sharedMemSize >>>
        (d_vec, (float4*) Pos, (float4*) Vel, soft2, numBodies);
    // d_vec is now the energy vector on device
    CUT_CHECK_ERROR("ENN Kernel execution failed");

// we speed this up with reduction parallel sum
// for parallel reduction sum kernel
    int sharedMemSize2 = p*sizeof(float); // needed for parallel reduction sum
    int numBlocks = numBodies/(2*p); // note 2
    numBlocks*=2;   // reduce6 is meant to run with more blocks
    dim3 grid2(numBlocks, 1, 1);

    float *d_odata;   // for sums on device
    CUDA_SAFE_CALL( cudaMalloc((void**) &d_odata, numBlocks*sizeof(float)) );
    float *h_odata = (float *) malloc(numBlocks*sizeof(float));    // sums on CPU

    // do the reduction sum each block, switch on blockSize
    switch(p) { // may need more cases here
        case 128:
         reduce6<128> <<< grid2, threads, sharedMemSize2>>>(d_vec, d_odata, numBodies);
         break;
        case 256:
         reduce6<256> <<< grid2, threads, sharedMemSize2>>>(d_vec, d_odata, numBodies);
         break;
        case 512:
         reduce6<512> <<< grid2, threads, sharedMemSize2>>>(d_vec, d_odata, numBodies);
         break;
    }
    CUT_CHECK_ERROR("reduce Kernel execution failed");
    CUDA_SAFE_CALL( cudaMemcpy( h_odata, d_odata, numBlocks*sizeof(float),
       cudaMemcpyDeviceToHost) );
    float gpu_result = 0.0;
    // sum the block sums
    for(int i=0; i<numBlocks; i++) gpu_result += h_odata[i];

// cleanup so far
    CUDA_SAFE_CALL(cudaFree(d_vec));
    CUDA_SAFE_CALL(cudaFree(d_odata));
    free(h_odata);

// if bary/helio extra momentum term needed 
    if (m0 >0.0){  
      callchangem0(Pos, 0.0);  // reset central mass back to zero
      float pxpypz[3] = {0.0,0.0,0.0};
      // call sum momenta routine
      sumMomenta(Pos,  Vel, p, numBodies, pxpypz);// returns sum of P
      float p2sum = pxpypz[0]*pxpypz[0]+pxpypz[1]*pxpypz[1]+pxpypz[2]*pxpypz[2];
      gpu_result += 1.0/(2.0*m0)*p2sum;  // add in the last needed term
    }
    
    return gpu_result;
}



///////////////////////////////////////////////////////
/* sum m*Vx m*Vy and m*Vy for all particles on GPU
   returned in pxpypz vector on host
   Pos and Vel are on device
   call GPU for m*v sum, not necessarily slow because in global mem
   call GPU for parallel reduction sum
   note central mass or V0 should be set to zero 
   before this is called
   This routine is needed for energy sum and for drift term computation
*/
///////////////////////////////////////////////////////
void sumMomenta(float* Pos, float* Vel, int p, int numBodies, float *pxpypz)
{
// for sum vector kernel
    int sharedMemSize = 0; // none needed for the sum vector
    dim3 threads(p,1,1); // number of threads per block is p  blockdim.x=p
    dim3 grid(numBodies/p, 1, 1);

// for parallel reduction sum kernel
    int sharedMemSize2 = p*sizeof(float); // needed for parallel reduction sum 
    int numBlocks = numBodies/(2*p); // note 2
    numBlocks*=2;   // reduce6 is meant to run with more blocks
    dim3 grid2(numBlocks, 1, 1);

    float* d_mv; // allocate mv vector on device
    CUDA_SAFE_CALL( cudaMalloc((void**) &d_mv, numBodies*sizeof(float)) );
    float *d_odata;   // for sums on device
    CUDA_SAFE_CALL( cudaMalloc((void**) &d_odata, numBlocks*sizeof(float)) );
    float *h_odata = (float *) malloc(numBlocks*sizeof(float));    // sums on CPU

// for x, y, z
    for(int coordpos=0;coordpos<3;coordpos++)
    {
      // compute the m*v vector
      computemv <<< grid, threads, sharedMemSize>>>(Pos, Vel, d_mv, coordpos);
      CUT_CHECK_ERROR("MV Kernel execution failed");

      // do the reduction sum each block, switch on blockSize
      switch(p) { // may need more cases here 
        case 128:
         reduce6<128> <<< grid2, threads, sharedMemSize2>>>(d_mv, d_odata, numBodies);
         break;
        case 256:
         reduce6<256> <<< grid2, threads, sharedMemSize2>>>(d_mv, d_odata, numBodies);
         break;
        case 512:
         reduce6<512> <<< grid2, threads, sharedMemSize2>>>(d_mv, d_odata, numBodies);
         break;
      }
      CUT_CHECK_ERROR("reduce Kernel execution failed");
   
      CUDA_SAFE_CALL( cudaMemcpy( h_odata, d_odata, numBlocks*sizeof(float), 
         cudaMemcpyDeviceToHost) );
      float gpu_result = 0.0;
      // sum the block sums
      for(int i=0; i<numBlocks; i++) gpu_result += h_odata[i];
      pxpypz[coordpos] = gpu_result;  // return the result in a vector

    }

// cleanup
    CUDA_SAFE_CALL(cudaFree(d_odata));
    CUDA_SAFE_CALL(cudaFree(d_mv));
    free(h_odata);

}


}
