/* similar to nbody_kernel except for total energy computation 
 */

#ifndef _ENN_KERNEL_H_
#define _ENN_KERNEL_H_

#include <math.h>
#include <stdio.h>


#define LOOP_UNROLL_E 4

// Macros to simplify shared memory addressing
#define SX(i) sharedPos[i]

///////////////////////////////////////////////
// soft2 is softening squared
// add potential energy term to ai from particle bj
// mass for bj is stored in bj.w
///////////////////////////////////////////////
__device__ float 
bodyBodyPE(float pei, float4 bi, float4 bj, float soft2) 
{
    float3 r;

    // r_ij  [3 FLOPS] 
    r.x = bj.x - bi.x;
    r.y = bj.y - bi.y;
    r.z = bj.z - bi.z;

    float distSqr = r.x * r.x + r.y * r.y + r.z * r.z;
    if (distSqr > 1e-12){ // leave off self term as it will be very large
       distSqr += soft2;
       float invDist = rsqrtf(distSqr);
       pei += bj.w * invDist;
    }

    return pei;
}

///////////////////////////////////////////////
// add forces from every particle looping through the blockdim
// This is the "tile_calculation" function from the GPUG3 article.
__device__ float tile_PE(float4 myPos, float pe, float soft2)
{
    extern __shared__ float4 sharedPos[];
    int i;
    // Here we unroll the loop
    for (i = 0; i < blockDim.x; ) 
    {
        pe= bodyBodyPE(pe, myPos, SX(i++), soft2);  // I think flipped in SDK?
#if LOOP_UNROLL_E > 1
        pe= bodyBodyPE(pe, myPos, SX(i++), soft2); 
#endif
#if LOOP_UNROLL_E > 2
        pe= bodyBodyPE(pe, myPos, SX(i++), soft2); 
        pe= bodyBodyPE(pe, myPos, SX(i++), soft2); 
#endif
#if LOOP_UNROLL_E > 4
        pe= bodyBodyPE(pe, myPos, SX(i++), soft2); 
        pe= bodyBodyPE(pe, myPos, SX(i++), soft2); 
        pe= bodyBodyPE(pe, myPos, SX(i++), soft2); 
        pe= bodyBodyPE(pe, myPos, SX(i++), soft2); 
#endif
    }
    return pe;
}

// WRAP is used to force each block to start working on a different 
// chunk (and wrap around back to the beginning of the array) so that
// not all multiprocessors try to read the same memory locations at 
// once.
#define WRAP(x,m) (((x)<m)?(x):(x-m))  // Mod without divide, works on values from 0 up to 2m

///////////////////////////////////////////////
__device__ float
computeBodyPE(float4 bodyPos, float4* positions, int numBodies, float soft2)
{
    extern __shared__ float4 sharedPos[];

    float pe= 0.0;
    
    int p = blockDim.x;
    int n = numBodies;
    int i,tile;
   // number of tiles is N/p = blockdim.x which is why wrap works

// load up mem for p particles then compute forces for all blocks with same memory

    for (i = 0, tile = 0; i < n; i += p, tile++)  // looping with step of tile (p) 
    { // load up shared memory 
        int idx = tile*blockDim.x + threadIdx.x;
        sharedPos[threadIdx.x] =  positions[idx];
//              positions[WRAP(blockIdx.x+tile, gridDim.x) * blockDim.x + threadIdx.x];
// no difference between these 2 seen
       
        __syncthreads();
        // This is the "tile_calculation" function from the GPUG3 article.
        pe= tile_PE(bodyPos, pe, soft2);
        __syncthreads();
    }


    return pe;
}

///////////////////////////////////////////////
// the thing you call, passing whole arrays of positions and velocities
// add the kinetic energies to the  potential energies
// return in a vector called enn
// if called for helio/bary coordinate system then m0 should not be zero
//   and Q0 should be set to zero
//   and an additional sum momentum square term should be added after this routine is called
///////////////////////////////////////////////

__global__ void BodyKEPE(float* enn, float4* Pos, float4* Vel, 
                float soft2, int numBodies)
{
    int index = __mul24(blockIdx.x,blockDim.x) + threadIdx.x;
    float4 myvel = Vel[index];
    float vel2 = (myvel.x*myvel.x + myvel.y*myvel.y + myvel.z*myvel.z); 
         // this is   v_i^2

    float4 mypos = Pos[index];   
    float pei=0.0;
    pei= computeBodyPE(mypos, Pos, numBodies, soft2); // Sum{j!=i} m_j/r_ij onto particle mypos
    enn[index] = 0.5*mypos.w*(vel2 - pei);
}

#endif // #ifndef _ENN_KERNEL_H_

